import dataclasses
from typing import Dict, Optional

import equinox as eqx
import haliax as hax
import haliax.nn as hnn
import jax.random as jrandom
from haliax.jax_utils import maybe_rng_split
from levanter.models.gpt2 import Gpt2Block, Gpt2Config, Gpt2Embeddings

from qkvflow.loop import checkpointed_while_loop


class WeightSharingTransformer(eqx.Module):

    config: Gpt2Config = eqx.field(static=True)
    block: Gpt2Block
    ln_f: hnn.LayerNorm

    num_check_points: int = eqx.field(static=True)

    @staticmethod
    def init(config: Gpt2Config, num_check_points, *, key):
        block = Gpt2Block.init(config, key=key)
        ln_f = hnn.LayerNorm.init(
            config.Embed, eps=config.layer_norm_epsilon, use_bias=config.use_bias
        )

        return WeightSharingTransformer(config, block, ln_f, num_check_points)

    def __call__(self, x, attn_mask, *, key=None):

        n_layers = self.config.num_layers

        def cond_fn(t_x):
            t, *_ = t_x
            return t < n_layers

        def body_fun(t_x):
            t, x = t_x
            key_t = None if key is None else jrandom.fold_in(key, t)
            output = self.block(x, mask=attn_mask, layer_idx=None, key=key_t)

            t = t + 1
            output = x + output

            return t, output

        output = checkpointed_while_loop(
            cond_fn,
            body_fun,
            init_val=(0, x),
            checkpoints=self.num_check_points,
        )

        x = output[1]
        x = self.ln_f(x)

        return x


class WeightSharingLMHeadModel(eqx.Module):
    transformer: WeightSharingTransformer
    embeddings: Gpt2Embeddings

    @property
    def config(self):
        return self.transformer.config

    @property
    def Vocab(self) -> hax.Axis:
        return self.embeddings.Vocab

    @property
    def Pos(self) -> hax.Axis:
        return self.config.Pos

    @classmethod
    def init(
        cls,
        Vocab: hax.Axis,
        config: Gpt2Config,
        num_check_points=5,
        *,
        key,
    ):
        k_t, k_embeddings = jrandom.split(key, 2)
        transformer = WeightSharingTransformer.init(
            config,
            num_check_points=num_check_points,
            key=k_t,
        )
        embeddings = Gpt2Embeddings.init(Vocab, config, key=k_embeddings)

        return WeightSharingLMHeadModel(transformer, embeddings)

    def __call__(self, input_ids, attn_mask=None, *, key=None):
        k_embed, k_transformer = maybe_rng_split(key, 2)
        x = self.embeddings.embed(input_ids, key=k_embed)
        x = self.transformer(x, attn_mask, key=k_transformer)
        lm_logits = self.embeddings.unembed(x)

        return lm_logits

    def compute_loss(
        self,
        example,
        *,
        key=None,
        reduction: Optional[hax.ReductionFunction] = hax.mean,
        reduction_axis: Optional[hax.AxisSelection] = None,
    ):
        logits = self(example.tokens, example.attn_mask, key=key)
        targets = hax.roll(example.tokens, -1, axis=self.Pos.name)
        target_y = hax.nn.one_hot(targets, self.Vocab, dtype=logits.dtype)
        return hnn.cross_entropy_loss(
            logits,
            self.Vocab,
            target_y,
            reduction,
            reduction_axis=reduction_axis,
            where=example.loss_mask,
        )

    @property
    def vocab_size(self) -> int:
        return self.Vocab.size

    def resize_vocab(self, new_size: int, key=None):
        new_embeddings = self.embeddings.resize_embeddings(new_size, key=key)
        return dataclasses.replace(self, embeddings=new_embeddings)

    def _state_dict_key_map(self) -> Dict[str, Optional[str]]:
        return {"transformer": None, "embeddings": None}
